#include "IrrlichtBase.h"
#include "ncbind/ncbind.hpp"

extern void message_log(const char* format, ...);
extern void error_log(const char *format, ...);

using namespace irr;
using namespace core;
using namespace video;
using namespace scene;
using namespace io;
using namespace gui;

/**
 * RXgN^
 */
IrrlichtBase::IrrlichtBase(iTJSDispatch2 *objthis)
	: objthis(objthis), device(NULL), attached(false),
	  eventMask(EMASK_ATTACH | EMASK_DETACH)
{
}

/**
 * fXgN^
 */
IrrlichtBase::~IrrlichtBase()
{
	stop();
	detach();
}

/**
 * hCȍ\
 */
void
IrrlichtBase::showDriverInfo()
{
	IVideoDriver *driver = device->getVideoDriver();
	if (driver) {
		dimension2d<s32> size = driver->getScreenSize();
		message_log("foCX̃XN[TCY:%d, %d", size.Width, size.Height);
		size = driver->getCurrentRenderTargetSize();
		message_log("foCXRenderTarget:%d, %d", size.Width, size.Height);
	}
}

/**
 * TJSCxgĂяoBȃIuWFNg̊Y\bhĂяoB
 * @param eventName Cxg
 */
void
IrrlichtBase::sendTJSEvent(const tjs_char *eventName)
{
	tTJSVariant method;
	if (TJS_SUCCEEDED(objthis->PropGet(0, eventName, NULL, &method, objthis))) { // Cxg\bh擾
		if (method.Type() == tvtObject) {
			iTJSDispatch2 *m = method.AsObjectNoAddRef();
			if (TJS_SUCCEEDED(m->IsInstanceOf(0, NULL, NULL, L"Function", m))) { // t@NVǂ
				tTJSVariant self(objthis, objthis);
				tTJSVariant *params[] = {&self};
				m->FuncCall(0, NULL, NULL, NULL, 1, params, method.AsObjectThisNoAddRef());
			}
		}
	}
}

/**
 * foCX̊蓖
 * @param hwnd eEChEnh
 * @param width obNobt@TCY
 * @param height obNobt@TCYc
 */
void
IrrlichtBase::attach(HWND hwnd, int width, int height)
{
	if (!attached && hwnd) {
		// foCX
		SIrrlichtCreationParameters params;
		params.WindowId     = reinterpret_cast<void*>(hwnd);
		params.DriverType    = EDT_DIRECT3D9;
		params.Stencilbuffer = true;
		params.Vsync = true;
		params.EventReceiver = this;
		params.AntiAlias = true;
		if (width != 0 && height != 0) {
			params.WindowSize = core::dimension2d<s32>(width, height);
		}
		if ((device = irr::createDeviceEx(params))) {
			TVPAddLog(L"IrrlichtfoCX");
			// eNX`̃ɂZeXgs悤ɁB
			// device->getSceneManager()->getParameters()->setAttribute(scene::ALLOW_ZWRITE_ON_TRANSPARENT, true);
			showDriverInfo();
			onAttach();
			if ((eventMask & EMASK_ATTACH)) {
				sendTJSEvent(L"onAttach");
			}
		} else {
			TVPThrowExceptionMessage(L"Irrlicht foCX̏Ɏs܂");
		}
		attached = true;
	}
}

/**
 * foCX̔j
 */
void
IrrlichtBase::detach()
{
	if (device) {
		if ((eventMask & EMASK_DETACH)) {
			sendTJSEvent(L"onDetach");
		}
		onDetach();
		device->drop();
		device = NULL;
	}
	attached = false;
}

/**
 * Irrlicht`揈
 * @param destRect `̈
 * @param srcRect `挳̈
 * @param destDC `DC
 * @return `悳ꂽ
 */
bool
IrrlichtBase::show(irr::core::rect<irr::s32> *destRect, irr::core::rect<irr::s32> *srcRect, HDC destDC)
{
	if (device) {
		// Ԃi߂ XXX tick O^ȂH
		device->getTimer()->tick();
		
		IVideoDriver *driver = device->getVideoDriver();
		// `Jn
		if (driver && driver->beginScene(true, true, irr::video::SColor(0,0,0,0))) {
			
			if ((eventMask & EMASK_BEFORE_SCENE)) {
				sendTJSEvent(L"onBeforeScene");
			}
			
			/// V[}l[W̕`
			ISceneManager *smgr = device->getSceneManager();
			if (smgr) {
				smgr->drawAll();
			}

			if ((eventMask & EMASK_AFTER_SCENE)) {
				sendTJSEvent(L"onAfterScene");
			}
			
			// ŗL
			update(driver);

			if ((eventMask & EMASK_BEFORE_GUI)) {
				sendTJSEvent(L"onBeforeGUI");
			}
			
			// GUI̕`
			IGUIEnvironment *gui = device->getGUIEnvironment();
			if (gui) {
				gui->drawAll();
			}

			if ((eventMask & EMASK_AFTER_GUI)) {
				sendTJSEvent(L"onAfterGUI");
			}
			
			// `抮
			driver->endScene(0, srcRect, destRect, destDC);
			return true;
		}
	}
	return false;
};

/**
 * Irrlicht ւ̃CxgM
 */
bool
IrrlichtBase::postEvent(SEvent &ev)
{
	if (device) {
		if (device->getGUIEnvironment()->postEventFromUser(ev) ||
			device->getSceneManager()->postEventFromUser(ev)) {
			return true;
		}
	}
	return false;
}

/**
 * Cxg
 * HWND w肵ĐĂ֌W Irrlicht g̓EChE
 * Cxg擾邱Ƃ͂ȂB̂ GUI Environment ̃Cxg
 * ɂ邱ƂɂȂB̓Kȃ\bhĂяo悤ɗvC XXX
 * @return  true
 */
bool
IrrlichtBase::OnEvent(const irr::SEvent &event)
{
	bool ret = false;
	if ((eventMask & EMASK_EVENT)) {
		tTJSVariant method;
		if (TJS_SUCCEEDED(objthis->PropGet(0, L"onEvent", NULL, &method, objthis))) { // Cxg\bh擾
			if (method.Type() == tvtObject) {
				iTJSDispatch2 *m = method.AsObjectNoAddRef();
				if (TJS_SUCCEEDED(m->IsInstanceOf(0, NULL, NULL, L"Function", m))) { // t@NVǂ
					tTJSVariant self(objthis, objthis);
					tTJSVariant ev;

					// SEvent ϊ
					typedef ncbInstanceAdaptor<SEvent> AdaptorT;
					iTJSDispatch2 *adpobj = AdaptorT::CreateAdaptor(new SEvent(event));
					if (adpobj) {
						ev = tTJSVariant(adpobj, adpobj);
						adpobj->Release();
					}

					tTJSVariant *params[] = {&self, &ev};
					tTJSVariant result;
					m->FuncCall(0, NULL, NULL, &result, 2, params, method.AsObjectThisNoAddRef());
					ret = (tjs_int)result != 0;
				}
			}
		}
	}
	return ret;
}

// --------------------------------------------------------------------------------

void
IrrlichtBase::start()
{
	stop();
	TVPAddContinuousEventHook(this);
}

void
IrrlichtBase::stop()
{
	TVPRemoveContinuousEventHook(this);
}
